<!DOCTYPE html>
<html>
    <head>
        <meta http-equiv="Content-type" content="text/html" charset="UTF-8" />
        <script>BEAR_ROUTE = '{{route}}';</script>
        <script src="{{asset:base-lib.js}}"></script>
    </head>
    <body>
        {{body}}
    </body>
</html>
